// hlavickove subory *******************************************************
// systemove funkcie
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// .h subor
#include "m3Dmath.h"

// vypocet velkosti vektora ************************************************
float m3Dvector::Magnitude(void)
{
 return(sqrt(x*x+y*y+z*z));
}

// normalizacia vektora ****************************************************
void m3Dvector::Normalize(void)
{
 M=Magnitude();

 if (M!=0.0f) { x/=M; y/=M; z/=M; }
}

// test kolizie kamery s trojuholnikom *************************************
bool m3Dmath::CameraInTriangle(m3Dvector triangle[],m3Dvector trianglenormal,m3Dvector line[],
                               m3Dvector triangleplanes[],float D1,float D2, float D3)
{
 // test kolizie pohyboveho vektora kamery s trojuholnikom
 if (!IntersectedTriangle(triangle,trianglenormal,line,&distance0)) return false;

 // vypocet priesecnika roviny, ktoru tvori trojuholnik a pohyboveho vektora kamery
 m3Dvector CameraTriangleIPoint=IntersectionPoint(trianglenormal,line,distance0);

 // test ci priesecnik patri trojuholniku, ak nie, nedoslo ku kolizii
 if (!PointInTriangle(CameraTriangleIPoint,triangleplanes,D1,D2,D3)) return false;

 // kamera kolidovala s trojuholnikom
 return true;
}

// vypocet vzdialenosti trojuholnika od pociatku suradnicovej sustavy ******
float m3Dmath::TriangleDistance(m3Dvector normal,m3Dvector point)
{
 // pouzije sa rovnica roviny (Ax+By+Cz+D=0), podla ktorej sa vypocita D=-(Ax+By+Cz)
 return -((normal.x*point.x)+(normal.y*point.y)+(normal.z*point.z));
}

// test kolizie pohyboveho vektora kamery s trojuholnikom *****************
bool m3Dmath::IntersectedTriangle(m3Dvector triangle[],
                                  m3Dvector trianglenormal,
                                  m3Dvector line[],
                                  float *dist)
{
 // vypocet vzdialenosti trojuholnika od pociatku suradnicovj sustavy
 // vypocitava sa D=-(Ax+By+Cz)
 // kde A,B,C=normala trojuholnika
 //     x,y,z=lubovolny vrchol tohto trojuholnika
 (*dist)=TriangleDistance(trianglenormal,triangle[0]);

 // vypocet vzdialenosti 1. bodu pohyboveho vektora kamery od trojuholnika (roviny, ktoru tvori)
 distance1=trianglenormal.x*line[0].x+ // Ax+
           trianglenormal.y*line[0].y+ // By+
           trianglenormal.z*line[0].z+ // Cz+
           (*dist);                    // D
 // vypocet vzdialenosti 2. bodu pohyboveho vektora kamery od trojuholnika (roviny, ktoru tvori)
 distance2=trianglenormal.x*line[1].x+ // Ax+
           trianglenormal.y*line[1].y+ // By+
           trianglenormal.z*line[1].z+ // Cz+
           (*dist);                    // D

 // INFO: sucin povodnej a novej vzdialenosti
 //fprintf(stderr,"distance1*distance2: %f\n",distance1*distance2);

 // vzajomne vynasovenie ziskanych vzdialenosti da kladny, resp. zaporny vysledok, co urci
 // vztah medzi povodnou a novou poziciou kamery 
 if (distance1*distance2>=0) return false;

 // pohybovy vektor kamery kolidoval s trojuholnikom (rovinou, ktoru tvori)
 return true;
}

// normalizacia vektora ****************************************************
m3Dvector m3Dmath::Normalize(m3Dvector vector)
{
 M=sqrt((vector.x*vector.x)+(vector.y*vector.y)+(vector.z*vector.z));

 if (M!=0.0f) { vector.x/=M; vector.y/=M; vector.z/=M; }

 return vector;
}

// skalarny sucin dvoch vektorov *******************************************
float m3Dmath::Dot(m3Dvector vector1,m3Dvector vector2)
{
 return((vector1.x*vector2.x)+(vector1.y*vector2.y)+(vector1.z*vector2.z));
}

// vzdialenost dvoch bodov *************************************************
float m3Dmath::Distance(m3Dvector pA,m3Dvector pB)
{
 return(sqrt((pB.x-pA.x)*(pB.x-pA.x)+(pB.y-pA.y)*(pB.y-pA.y)+(pB.z-pA.z)*(pB.z-pA.z)));
}

// vypocet priesecnika roviny, ktoru tvori trojuholnik a pohyboveho vektora kamery
m3Dvector m3Dmath::IntersectionPoint(m3Dvector trianglenormal,m3Dvector line[],float dist)
{
 // vypocet hodnot pohyboveho vektora
 m_vector=line[1]-line[0];
 m_vector=Normalize(m_vector);
 // ziskanie vzdilenosti 1. bodu pohyboveho vektora od roviny
 Numerator=-(trianglenormal.x*line[0].x+
             trianglenormal.y*line[0].y+
             trianglenormal.z*line[0].z+
             dist);
 // skalarny sucin (dot product) dvoch vektorov
 Denominator=Dot(trianglenormal,m_vector);

 // v pripade, ze pohybovy vektor je kolmy na normalu trojuholnika, vrati sa lubovolny
 // bod pohyboveho vektora
 if (Denominator==0.0f) return line[0];

 D=Numerator/Denominator;

 m_point.x=(line[0].x+(m_vector.x*D));
 m_point.y=(line[0].y+(m_vector.y*D));
 m_point.z=(line[0].z+(m_vector.z*D));

 // vrati sa priesecnik pohyboveho vektora kamery s rovinou tvorenou trojuholnikom
 return m_point;
}

// test ci priesecnik patri trojuholniku, ak nie, nedoslo ku kolizii *******
bool m3Dmath::PointInTriangle(m3Dvector IPoint,m3Dvector triangleplanes[],float D1,float D2,float D3)
{
 // ak bude priesecnik aspon za jednou stranou, nie je v trojuholniku
 if (IPoint.x*triangleplanes[0].x+IPoint.y*triangleplanes[0].y+IPoint.z*triangleplanes[0].z+D1>0.0f) return false;
 if (IPoint.x*triangleplanes[1].x+IPoint.y*triangleplanes[1].y+IPoint.z*triangleplanes[1].z+D2>0.0f) return false;
 if (IPoint.x*triangleplanes[2].x+IPoint.y*triangleplanes[2].y+IPoint.z*triangleplanes[2].z+D3>0.0f) return false;

 // priesecnik je pred vsetkymi troma stranami => je vnutri v trojuholniku
 return true;
}

// vrati absolutnu hodnotu cisla *******************************************
float m3Dmath::myAbs(float N)
{
 if (N<0.0f) return (0.0f-N);

 return N;
}

// test kolizie kamery s trojuholnikom - s vyuzitim gule *******************
bool m3Dmath::CameraInTriangleS(m3Dvector triangle[],m3Dvector trianglenormal,m3Dvector center,float radius,
                                m3Dvector triangleplanes[],float D1,float D2, float D3)
{
 // test kolizie gule s trojuholnikom, zaroven sa vypocitaju suradnice kolizneho bodu
 if (IntersectedSphere(triangle,trianglenormal,center,radius,&distance0)) {
  if (PointInTriangle(m_point,triangleplanes,D1,D2,D3)) {
   // posun kamery v opacnom smere (efekt klzania po koliznej ploche)
   m_vector=trianglenormal*radius;
   ret_pos=m_point+m_vector;
   // kamera kolidovala s trojuholnikom
   return true;
  }

  if (EdgeCollision(center,triangle,radius)) {
   // posun kamery v opacnom smere (efekt klzania po koliznej hrane)
   m_vector=Normalize(center-m_point)*radius;
   ret_pos=(m_point+m_vector);
   // kamera kolidovala s hranou trojuholnika
   return true;
  }
 }

 return false;
}

// test kolizie gule s trojuholnikom ***************************************
bool m3Dmath::IntersectedSphere(m3Dvector triangle[],
                                m3Dvector trianglenormal,
                                m3Dvector center,float radius,
                                float *dist)
{
 // vypocet vzdialenosti trojuholnika od pociatku suradnicovj sustavy
 // vypocitava sa D=-(Ax+By+Cz)
 // kde A,B,C=normala trojuholnika
 //     x,y,z=lubovolny vrchol tohto trojuholnika
 (*dist)=TriangleDistance(trianglenormal,triangle[0]);

 // vypocet vzdialenosti stredu gule od trojuholnika (roviny, ktoru tvori)
 distance1=trianglenormal.x*center.x+ // Ax+
           trianglenormal.y*center.y+ // By+
           trianglenormal.z*center.z+ // Cz+
           (*dist);                   // D

 // INFO: vzdielnost stredu gule od roviny
 //fprintf(stderr,"distance1: %f\n",distance1);

 // ak je vzdialenost gule od roviny vacsia ako jej polomer, nedoslo ku kolizii
 if (myAbs(distance1)>=radius) return false;

 // presun stredu gule do roviny, ktoru tvori trojuholnik
 m_point=center-(trianglenormal*distance1);

 // gula kolidovala s trojuholnikom (rovinou, ktoru tvori)
 return true;
}

// test kolizie gule s hranami trojuholnika ********************************
bool m3Dmath::EdgeCollision(m3Dvector center,m3Dvector triangle[],float radius)
{
 // 1. hrana
 m_point=PointOnLine(triangle[0],triangle[1],center);
 if (Distance(m_point,center)<radius) return true;
 // 2. hrana
 m_point=PointOnLine(triangle[1],triangle[2],center);
 if (Distance(m_point,center)<radius) return true;
 // 3. hrana
 m_point=PointOnLine(triangle[2],triangle[0],center);
 if (Distance(m_point,center)<radius) return true;
 // ku kolizii nedoslo
 return false;
}

// vrati bod na ciare pA-pB, ktory je najblizsie k bodu pC *****************
m3Dvector m3Dmath::PointOnLine(m3Dvector pA,m3Dvector pB,m3Dvector pC)
{
 m_vectA=pC-pA;
 m_vectB=Normalize(pB-pA);

 D=Distance(pA,pB);

 // projekcia vektora m_vect1 do vektora m_vect2
 // ziskanie vzdialenosti projektovaneho vektoru od bodu pA
 M=Dot(m_vectB,m_vectA);

 // vzdialenost je mensia ako pA
 if (M<=0) return pA;

 // vzdialenost je vacsia ako pB
 if (M>=D) return pB;

 // vypocet vektora dlzky M v smere vektora m_vect2
 m_vector=m_vectB*M;

 // posun bodu pA pomocou vektora m_vect2
 return(pA+m_vector);
}

/*
  pri tvorbe tohto kodu som sa inspiroval zverejnenym kodom autora:

  Ben Humphrey (DigiBen)
  Game Programmer
  DigiBen@GameTutorials.com
  Co-Web Host of www.GameTutorials.com
  date: 1/23/02
*/
