// hlavickove subory *******************************************************
// systemove funkcie
#include <math.h>
#include <stdio.h>
// GL a GLU kniznica
#include <GL/gl.h>
#include <GL/glu.h>

// .h subor
#include "mFrustum.h"

// zoznam stran Frustum
enum FrustumSide
{
 RIGHT =0, // prava strana Frustum
 LEFT  =1, // lava strana Frustum
 BOTTOM=2, // spodna strana Frustum
 TOP   =3, // horna strana Frustum
 BACK  =4, // zadna strana Frustum
 FRONT =5  // predna strana Frustum
};

// zoznam hodnot stran Frustum
enum PlaneData
{
 A=0, // X hodnota normaly
 B=1, // Y hodnota normaly
 C=2, // Z hodnota normaly
 D=3  // vzdialenost strany od pociatku
};

// konstruktor *************************************************************
FrustumCulling::FrustumCulling()
{

}

// destruktor **************************************************************
FrustumCulling::~FrustumCulling()
{

}

// normalizacia strany Frustum *********************************************
void FrustumCulling::NormalizePlane(GLfloat frustum[6][4],GLuint side)
{
 // vypocet magnitudy normaly strany
 // magnitude = sqrt( x^2 + y^2 + z^2)
 magnitude=(GLfloat) sqrt(frustum[side][A]*frustum[side][A]+
                          frustum[side][B]*frustum[side][B]+
                          frustum[side][C]*frustum[side][C]);

 // delenie hodnot strany magnitudou jej normaly
 frustum[side][A]/=magnitude;
 frustum[side][B]/=magnitude;
 frustum[side][C]/=magnitude;
 frustum[side][D]/=magnitude;
}

// funkcia pre vypocet Frustum *********************************************
void FrustumCulling::FrustumUpdate(void)
{
 // ziskanie projekcnej matice
 glGetFloatv(GL_PROJECTION_MATRIX,proj);

 // nacitanie modelview matice
 glGetFloatv(GL_MODELVIEW_MATRIX,modl);

 // INFO: vypis obsahu projekcnej matice
 //fprintf(stderr,"%f %f %f %f\n",proj[0],proj[1],proj[2],proj[3]); //x,0,0,0
 //fprintf(stderr,"%f %f %f %f\n",proj[4],proj[5],proj[6],proj[7]); //0,x,0,0
 //fprintf(stderr,"%f %f %f %f\n",proj[8],proj[9],proj[10],proj[11]); //0,0,x,x
 //fprintf(stderr,"%f %f %f %f\n",proj[12],proj[13],proj[14],proj[15]); //0,0,x,0

 // v pripade, ze sa nemeni projekcna matica, mozno vypocet koeficientov zjednodusit
 // ziskanie orezavacich rovin kombinaciou matic
 clip[ 0]=modl[ 0]*proj[ 0];
 clip[ 1]=modl[ 1]*proj[ 5];
 clip[ 2]=modl[ 2]*proj[10]+modl[ 3]*proj[14];
 clip[ 3]=modl[ 2]*proj[11];

 clip[ 4]=modl[ 4]*proj[ 0];
 clip[ 5]=modl[ 5]*proj[ 5];
 clip[ 6]=modl[ 6]*proj[10]+modl[ 7]*proj[14];
 clip[ 7]=modl[ 6]*proj[11];

 clip[ 8]=modl[ 8]*proj[ 0];
 clip[ 9]=modl[ 9]*proj[ 5];
 clip[10]=modl[10]*proj[10]+modl[11]*proj[14];
 clip[11]=modl[10]*proj[11];

 clip[12]=modl[12]*proj[ 0];
 clip[13]=modl[13]*proj[ 5];
 clip[14]=modl[14]*proj[10]+modl[15]*proj[14];
 clip[15]=modl[14]*proj[11];
 /*
 // ziskanie orezavacich rovin kombinaciou matic
 clip[ 0]=modl[ 0]*proj[ 0]+modl[ 1]*proj[ 4]+modl[ 2]*proj[ 8]+modl[ 3]*proj[12];
 clip[ 1]=modl[ 0]*proj[ 1]+modl[ 1]*proj[ 5]+modl[ 2]*proj[ 9]+modl[ 3]*proj[13];
 clip[ 2]=modl[ 0]*proj[ 2]+modl[ 1]*proj[ 6]+modl[ 2]*proj[10]+modl[ 3]*proj[14];
 clip[ 3]=modl[ 0]*proj[ 3]+modl[ 1]*proj[ 7]+modl[ 2]*proj[11]+modl[ 3]*proj[15];

 clip[ 4]=modl[ 4]*proj[ 0]+modl[ 5]*proj[ 4]+modl[ 6]*proj[ 8]+modl[ 7]*proj[12];
 clip[ 5]=modl[ 4]*proj[ 1]+modl[ 5]*proj[ 5]+modl[ 6]*proj[ 9]+modl[ 7]*proj[13];
 clip[ 6]=modl[ 4]*proj[ 2]+modl[ 5]*proj[ 6]+modl[ 6]*proj[10]+modl[ 7]*proj[14];
 clip[ 7]=modl[ 4]*proj[ 3]+modl[ 5]*proj[ 7]+modl[ 6]*proj[11]+modl[ 7]*proj[15];

 clip[ 8]=modl[ 8]*proj[ 0]+modl[ 9]*proj[ 4]+modl[10]*proj[ 8]+modl[11]*proj[12];
 clip[ 9]=modl[ 8]*proj[ 1]+modl[ 9]*proj[ 5]+modl[10]*proj[ 9]+modl[11]*proj[13];
 clip[10]=modl[ 8]*proj[ 2]+modl[ 9]*proj[ 6]+modl[10]*proj[10]+modl[11]*proj[14];
 clip[11]=modl[ 8]*proj[ 3]+modl[ 9]*proj[ 7]+modl[10]*proj[11]+modl[11]*proj[15];

 clip[12]=modl[12]*proj[ 0]+modl[13]*proj[ 4]+modl[14]*proj[ 8]+modl[15]*proj[12];
 clip[13]=modl[12]*proj[ 1]+modl[13]*proj[ 5]+modl[14]*proj[ 9]+modl[15]*proj[13];
 clip[14]=modl[12]*proj[ 2]+modl[13]*proj[ 6]+modl[14]*proj[10]+modl[15]*proj[14];
 clip[15]=modl[12]*proj[ 3]+modl[13]*proj[ 7]+modl[14]*proj[11]+modl[15]*proj[15];
 */
 // ziskanie stran orezavacich rovin
 m_Frustum[RIGHT][A]=clip[ 3]-clip[ 0];
 m_Frustum[RIGHT][B]=clip[ 7]-clip[ 4];
 m_Frustum[RIGHT][C]=clip[11]-clip[ 8];
 m_Frustum[RIGHT][D]=clip[15]-clip[12];
 // normalizujeme stranu zo ziskanej normaly (A,B,C) a vzdialenosti (D)
 NormalizePlane(m_Frustum,RIGHT);

 m_Frustum[LEFT][A]=clip[ 3]+clip[ 0];
 m_Frustum[LEFT][B]=clip[ 7]+clip[ 4];
 m_Frustum[LEFT][C]=clip[11]+clip[ 8];
 m_Frustum[LEFT][D]=clip[15]+clip[12];
 NormalizePlane(m_Frustum,LEFT);

 m_Frustum[BOTTOM][A]=clip[ 3]+clip[ 1];
 m_Frustum[BOTTOM][B]=clip[ 7]+clip[ 5];
 m_Frustum[BOTTOM][C]=clip[11]+clip[ 9];
 m_Frustum[BOTTOM][D]=clip[15]+clip[13];
 NormalizePlane(m_Frustum,BOTTOM);

 m_Frustum[TOP][A]=clip[ 3]-clip[ 1];
 m_Frustum[TOP][B]=clip[ 7]-clip[ 5];
 m_Frustum[TOP][C]=clip[11]-clip[ 9];
 m_Frustum[TOP][D]=clip[15]-clip[13];
 NormalizePlane(m_Frustum,TOP);

 m_Frustum[BACK][A]=clip[ 3]-clip[ 2];
 m_Frustum[BACK][B]=clip[ 7]-clip[ 6];
 m_Frustum[BACK][C]=clip[11]-clip[10];
 m_Frustum[BACK][D]=clip[15]-clip[14];
 NormalizePlane(m_Frustum,BACK);

 m_Frustum[FRONT][A]=clip[ 3]+clip[ 2];
 m_Frustum[FRONT][B]=clip[ 7]+clip[ 6];
 m_Frustum[FRONT][C]=clip[11]+clip[10];
 m_Frustum[FRONT][D]=clip[15]+clip[14];
 NormalizePlane(m_Frustum,FRONT);
}

/*
  pre test, ci je bod, gula, kocka... vnutri Frustum pouzivame predpoklad,
  ze vsetky normaly stran Frustum smeruju dovnutra Frustum priestoru a teda
  pokial je bod pred vsetkymi 6-timi stranami, je vnutri Frustum
*/

// test, ci je bod vo Frustum **********************************************
bool FrustumCulling::PointInFrustum(GLfloat x,GLfloat y,GLfloat z)
{
 // test vsetkych 6-tich stran
 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*x+m_Frustum[i][B]*y+m_Frustum[i][C]*z+m_Frustum[i][D]<=0) {
   // bod je za stranou => je vonku z Frustum
   return false;
  }
 }
 // bod je pred vsetkymi 6-timi stranami => je vo Frustum
 return true;
}

// test, ci je trojuholnik vo Frustum **************************************
bool FrustumCulling::TriangleInFrustum(GLfloat v1,GLfloat v2,GLfloat v3,
                                       GLfloat v4,GLfloat v5,GLfloat v6,
                                       GLfloat v7,GLfloat v8,GLfloat v9)
{
 // test vsetkych 6-tich stran frustum
 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*v1+m_Frustum[i][B]*v2+m_Frustum[i][C]*v3+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*v4+m_Frustum[i][B]*v5+m_Frustum[i][C]*v6+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*v7+m_Frustum[i][B]*v8+m_Frustum[i][C]*v9+m_Frustum[i][D]>0)
   continue;

  return false; // trojuholnik sa vyluci iba ak lezia vsetky jeho 3 vrcholy
                // za niektorou z rezovych rovin
 }
 // aspon 1 vrchol trojuholniku je pred niektorou zo stran => trojuholnik je vo Frustum
 return true;
}

// test, ci je gula vo Frustum *********************************************
bool FrustumCulling::SphereInFrustum(GLfloat x,GLfloat y,GLfloat z,GLfloat radius)
{
 // test vsetkych 6-tich stran
 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*x+m_Frustum[i][B]*y+m_Frustum[i][C]*z+m_Frustum[i][D]<=-radius) {
   // stred gule+jej polomer je za stranou => je vonku z Frustum
   return false;
  }
 }
 // gula je pred vsetkymi 6-timi stranami => je vo Frustum
 return true;
}

// test, ci je kocka vo Frustum ********************************************
bool FrustumCulling::CubeInFrustum(GLfloat x,GLfloat y,GLfloat z,GLfloat size)
{
 // test vsetkych 6-tich stran
 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z-size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y-size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x-size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*(x+size)+m_Frustum[i][B]*(y+size)+m_Frustum[i][C]*(z+size)+m_Frustum[i][D]>0)
   continue;
  // stred kocky+1/2 jej velkosti je za stranou => je vonku z Frustum
  return false;
 }
 // kocka je pred vsetkymi 6-timi stranami => je vo Frustum
 return true;
}

// test, ci je kocka vo Frustum ********************************************
bool FrustumCulling::CubeInFrustum2(GLfloat x1,GLfloat x2,
                                    GLfloat y1,GLfloat y2,
                                    GLfloat z1,GLfloat z2)
{
 // INFO: vypis vrcholov kocky
 //fprintf(stderr,"x1:%f, x2:%f, y1:%f, y2:%f, z1:%f, z2:%f\n",x1,x2,y1,y2,z1,z2);

 // test vsetkych 6-tich stran
 for (i=0;i<6;i++) {
  if (m_Frustum[i][A]*x1+m_Frustum[i][B]*y1+m_Frustum[i][C]*z1+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*x2+m_Frustum[i][B]*y1+m_Frustum[i][C]*z1+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*x2+m_Frustum[i][B]*y2+m_Frustum[i][C]*z1+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*x1+m_Frustum[i][B]*y2+m_Frustum[i][C]*z1+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*x1+m_Frustum[i][B]*y1+m_Frustum[i][C]*z2+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*x2+m_Frustum[i][B]*y1+m_Frustum[i][C]*z2+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*x2+m_Frustum[i][B]*y2+m_Frustum[i][C]*z2+m_Frustum[i][D]>0)
   continue;
  if (m_Frustum[i][A]*x1+m_Frustum[i][B]*y2+m_Frustum[i][C]*z2+m_Frustum[i][D]>0)
   continue;

  return false; // kocka sa vyluci iba ak lezi vsetkych jej 8 vrcholov
                // za niektorou z rezovych rovin
 }

 return true; // ak nelezi cela kocka za aspon 1 rezovou rovinou, musi sa kreslit
}

/*
  tento kod vznikol po uprave zverejneneho kodu autora:

  Ben Humphrey (DigiBen)
  Game Programmer
  DigiBen@GameTutorials.com
  Co-Web Host of www.GameTutorials.com
  date: 11/26/01
*/
