// *************************************************************************
// * autor:   Marek Sopko                                                  *
// * kontakt: ... Trencin, Slovakia                                        *
// *          http:   www.avatar.host.sk                                   *
// *          e-mail: avatar&host.sk                                       *
// * popis aplikacie: OpenGL efekty                                        *
// *************************************************************************

// hlavickove subory *******************************************************
// systemove funkcie
#include <stdio.h>
#include <stdlib.h>
// GL a GLU kniznica (OpenGL)
#include <GL/gl.h>
//#include <GL/glu.h>
// funkcie SDL kniznice
#include <SDL/SDL.h>

// vlastne funkcie pre pracu s OpenGL - hlavny interface aplikacie
#include "mOGL.h"

// .h subor
#include "priklad15.h"

/***************************************************************************
 * deklaracie premennych a konstant                                        *
 ***************************************************************************/
// sirka, vyska a farebna hlbka (pocet bitov na 1 pixel) SDL okna
#define SCR_WIDTH  720
#define SCR_HEIGHT 486
#define SCR_BPP     32
// pozn.: v X11, VidMode nemeni rozlisenie obrazovky, vo Win32,
// ChangeDisplaySettings moze zmenit BPP

// meno aplikacie
const char app_title[24]="OpenGL GFX";
// SDL potrebne premenne
const SDL_VideoInfo *videoInfo=NULL; // informacie o obrazovke
int videoFlags=0;                    // znacky pre SDL_SetVideoMode
SDL_Event event;                     // obsluha udalosti
SDL_Surface *surface;                // SDL kresliaca plocha
Uint8 *keystate;                     // pole SDL stlacanych klaves
Uint8  mousestate;                   // SDL mys

// instancia triedy mOGL - hlavny interface aplikacie
mOGL *m_OGL=NULL;

// cislo pouzitej OpenGL kamery
unsigned int Cnum=0;

// pomocne premenne
int a=0,b=0,c=0;

/***************************************************************************
 * definicie funkcii aplikacie                                             *
 ***************************************************************************/
// uvolnenie alokovanych zdrojov a obnovenie povodnych nastaveni obrazovky *
void APP_quit(int returnCode)
{
 if (m_OGL!=NULL) delete m_OGL; // zrusenie instancie triedy mOGL
 m_OGL=NULL;                    // nastavenie m_OGL=NULL

 SDL_Quit();                    // zrusenie SDL okna

 exit(returnCode);              // ukoncenie aplikacie
}

// modifikacia nastaveni kresliacej plochy po zmene velkosti okna **********
int WND_resize(int width,int height)
{
 // zmena SDL kresliacej plochy (okna)
 surface=SDL_SetVideoMode(width,height,SCR_BPP,videoFlags);
 if (!surface) {
  fprintf(stderr,"Chyba nastavenia SDL kresliacej plochy po zmene velkosti okna: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // reinicializacia nastaveni kamery po zmene velkosti okna
 if (m_OGL!=NULL) m_OGL->m_Camera[Cnum].cam_reinit((GLfloat) width,(GLfloat) height);

 // umiestnenie kurzoru mysi na stred okna
 SDL_WarpMouse(SCR_WIDTH/2,SCR_HEIGHT/2);
 // este raz, po starte tak nedojde k nechcenej rotacii kamery
 SDL_WarpMouse(SCR_WIDTH/2,SCR_HEIGHT/2);

 return(0);
}

// obhospodarenie vstupu z klavesnice a mysi *******************************
void APP_input(void)
{
 // nacitanie zmeny polohy mysi
 mousestate=SDL_GetMouseState(&b,&a);
 // umiestnenie kurzoru mysi na stred okna
 SDL_WarpMouse(SCR_WIDTH/2,SCR_HEIGHT/2);
 // uprava posunu podla velkosti okna
 a-=SCR_HEIGHT/2; b-=SCR_WIDTH/2;
 // stlacenie laveho a praveho tlacidla mysi
 m_OGL->m_Camera[Cnum].isFire1=SDL_GetMouseState(NULL,NULL)&SDL_BUTTON(SDL_BUTTON_LEFT);
 m_OGL->m_Camera[Cnum].isFire2=SDL_GetMouseState(NULL,NULL)&SDL_BUTTON(SDL_BUTTON_RIGHT);

 // nacitanie stlacenych klaves
 keystate=SDL_GetKeyState(NULL);
 if (keystate[SDLK_ESCAPE]) APP_quit(0);               // ESC
 m_OGL->m_Camera[Cnum].goForward=keystate[SDLK_UP];    // sipka hore
 m_OGL->m_Camera[Cnum].goBack=   keystate[SDLK_DOWN];  // sipka dole
 m_OGL->m_Camera[Cnum].goRight=  keystate[SDLK_RIGHT]; // sipka doprava
 m_OGL->m_Camera[Cnum].goLeft=   keystate[SDLK_LEFT];  // sipka dolava
 m_OGL->m_Camera[Cnum].goUp=     keystate[SDLK_a];     // posun kamery hore
 m_OGL->m_Camera[Cnum].goDown=   keystate[SDLK_z];     // posun kamery dole
 m_OGL->m_Camera[Cnum].goJump=   keystate[SDLK_j];     // skok
                                                       // chodza alebo beh
 if (keystate[SDLK_RSHIFT]||keystate[SDLK_LSHIFT]) m_OGL->m_Camera[Cnum].goFast=1;

 c=0;
 if (keystate[SDLK_q]) c=-10;                          // rotacia kamery vlavo
 else if (keystate[SDLK_w]) c=10;                      // rotacia kamery vpravo

 if (keystate[SDLK_r]) m_OGL->m_Camera[Cnum].cam_reset();      // reset pohladu kamery
 if (keystate[SDLK_c]) m_OGL->m_Camera[Cnum].CollisionOnOff(); // zap/vyp detekcie kolizii
 if (keystate[SDLK_m]) m_OGL->change_PolygonMode();   // zmena sposobu zobrazenia objektov
 if (keystate[SDLK_l]) m_OGL->change_LightingOnOff(); // zap/vyp osvetlenia
 if (keystate[SDLK_n]) m_OGL->change_NormalsOnOff();  // zap/vyp vykreslovania normal
 if (keystate[SDLK_g]) m_OGL->change_FogOnOff();      // zap/vyp hmly
 if (keystate[SDLK_f]) m_OGL->change_FogMode();       // zmena parametrov hmly
 if (keystate[SDLK_t]) m_OGL->change_TexOnOff();      // zap/vyp texturovania
 if (keystate[SDLK_h]) m_OGL->change_HelpOnOff();     // zap/vyp vypisu helpu

 // rotacia kamery
 m_OGL->m_Camera[Cnum].cam_rotate((GLfloat) a,(GLfloat) b,(GLfloat) c);
 // posun kamery
 m_OGL->m_Camera[Cnum].cam_move();
 // aktualizacia parametrov kamery
 m_OGL->m_Camera[Cnum].cam_update();
}

// inicializacia SDL video subsystemu **************************************
int SDL_initialize(void)
{
 // inicializacia SDL video subsystemu
 if (SDL_Init(SDL_INIT_VIDEO)<0) {
  fprintf(stderr,"Chyba inicializacie SDL video subsystemu: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // ziskanie informacii o SDL video subsysteme
 videoInfo=SDL_GetVideoInfo();
 if (!videoInfo) {
  fprintf(stderr,"Chyba ziskania informacii o SDL video subsysteme: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // nastavenie atributov SDL kresliacej plochy
 // 8 bitov pre cervenu, zelenu a modru farbu
 // 16 bitovy zasobnik hlbky
 // 1 pre pouzitie dvojiteho zasobnika pre prekreslovanie
 // 1 bitovy stencil zasobnik
 SDL_GL_SetAttribute(SDL_GL_RED_SIZE,    8);
 SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,  8);
 SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,   8);
 //SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,  8);
 SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,  16);
 SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
 SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE,1);
 // pozn.: SDL_DOUBLEBUFFER sa pouziva v SDL_GL_SetAttribute,
 // v SDL_SetVideoMode to ovplivni iba 2D prekreslovanie

 // znacky pre SDL_SetVideoMode
 videoFlags =SDL_OPENGL;          // zapne OpenGL v SDL
 videoFlags|=SDL_GL_DOUBLEBUFFER; // alokuje dvojity zasobnik pre prekreslovanie
 videoFlags|=SDL_HWPALETTE;       // ulozi paletu do HW
 videoFlags|=SDL_RESIZABLE;       // umozni menit velkost SDL okna

 // kontrola ci mozno SDL kresliacu plochu umiestnit do HW
 if (videoInfo->hw_available) videoFlags|=SDL_HWSURFACE;
 else videoFlags|=SDL_SWSURFACE;

 // kontrola, ci mozno pracovat s HW blit
 if (videoInfo->blit_hw) videoFlags|=SDL_HWACCEL;

 // zapnutie FSAA ... DOROBIT! zistenie glext
 //SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS,1);
 //SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES,2);

 // nastavenie SDL kresliacej plochy
 surface=SDL_SetVideoMode(SCR_WIDTH,SCR_HEIGHT,SCR_BPP,videoFlags);
 if (!surface) {
  fprintf(stderr,"Chyba nastavenia SDL kresliacej plochy: %s\n",SDL_GetError());
  APP_quit(1);
 }

 // nastavenie nazvu okna a ikony
 SDL_WM_SetCaption(app_title,app_title);

 return(0);
}

// funkcia pre kreslenie OpenGL sceny **************************************
int OGL_SDL_draw_scene(void)
{
 // vykreslenie OpenGL sceny
 m_OGL->gl_draw();

 // vymena zobrazovacich zasobnikov, zobrazenie sceny
 SDL_GL_SwapBuffers();

 return(0);
}

// hlavna vykonavacia funkcia aplikacie ************************************
int main(int argc,char **argv)
{
 int isDone=0;   // premenna hlavneho cyklu
 int isActive=1; // SDL okno je/nie je aktivne

 m_OGL=NULL; // tvorba instancie triedy COGL
 if (m_OGL==NULL) {
  m_OGL=new mOGL();
  if (m_OGL==NULL) {
   fprintf(stderr,"Chyba tvorby instancie triedy mOGL!\n");
   APP_quit(1);
  }
 }

 // inicializacia SDL
 SDL_initialize();
 // inicializacia OpenGL 0=OK, 1=ERROR
 if (m_OGL->gl_initialize()) APP_quit(1);

 WND_resize(SCR_WIDTH,SCR_HEIGHT); // upravenie velkosti SDL okna

 SDL_ShowCursor(SDL_DISABLE); // schovanie kurzoru mysi

 // hlavny programovy cyklus
 while (!isDone) {
  // ukladanie a spracovavanie udalosti
  while (SDL_PollEvent(&event)) {
   switch (event.type) {
    case SDL_ACTIVEEVENT:    // kreslenie iba ak je okno aktivne
     if (event.active.gain==0) isActive=0;
     else isActive=1;
     break;
    case SDL_VIDEORESIZE:    // zmena velkosti okna
     WND_resize(event.resize.w,event.resize.h);
     break;
    case SDL_QUIT: isDone=1; // zrusenie SDL okna
     break;
    default: break;
   }
  }

  // obsluha klavesnice a mysi
  APP_input();

  // kreslenie sceny
  if (isActive) OGL_SDL_draw_scene();
 }

 // ukoncenie aplikacie
 APP_quit(0);

 return(0);
}
